import glob
import json
import os
import shutil
import sys
import tempfile
import unittest
import uuid

import pygit2

from . import commit, config, manage


class TestSetupReporter(unittest.TestCase):
    def setUp(self):
        self.repository = tempfile.mkdtemp()
        pygit2.init_repository(self.repository, False)

        self.fixtures_dir = os.path.join(os.path.dirname(__file__), "fixtures")

        script_basename = "a_script.py"
        script_file_fixture = os.path.join(self.fixtures_dir, script_basename)
        self.script_file = os.path.join(self.repository, script_basename)
        shutil.copyfile(script_file_fixture, self.script_file)

        package_basename = "a_package"
        package_dir_fixture = os.path.join(self.fixtures_dir, package_basename)
        self.package_dir = os.path.join(self.repository, package_basename)
        shutil.copytree(package_dir_fixture, self.package_dir)

        package_files = [
            os.path.relpath(python_file, start=self.repository)
            for python_file in glob.glob(os.path.join(self.package_dir, "*.py"))
        ]
        commit.commit_files(
            self.repository,
            "refs/heads/master",
            [script_basename, *package_files],
            "initial commit",
        )

        self.reporter_token = str(uuid.uuid4())

    def tearDown(self) -> None:
        DEBUG = os.getenv("DEBUG")
        if DEBUG != "1":
            shutil.rmtree(self.repository)
        else:
            print(
                f"DEBUG=1: Retaining test directory - {self.repository}",
                file=sys.stderr,
            )

    def test_add_reporter_for_package(self):
        config.initialize(
            self.repository,
            self.package_dir,
            "a_package",
            reporter_token=self.reporter_token,
        )
        self.assertTrue(
            os.path.exists(os.path.join(self.repository, config.CONFIG_FILENAME))
        )

        manage.add_reporter(self.repository, self.package_dir)

        infestor_json_path = config.default_config_file(self.repository)
        self.assertTrue(os.path.exists(infestor_json_path))
        with open(infestor_json_path, "r") as ifp:
            infestor_json = json.load(ifp)

        expected_config_json = {
            "a_package": {
                "python_root": "a_package",
                "project_name": "a_package",
                "relative_imports": False,
                "reporter_token": self.reporter_token,
                "reporter_filepath": "report.py",
            }
        }
        self.assertDictEqual(infestor_json, expected_config_json)

        self.assertTrue(
            os.path.exists(
                os.path.join(
                    self.repository,
                    self.package_dir,
                    "report.py",
                )
            )
        )

    def test_list_system_reports_for_package_with_no_system_reports(self):
        config.initialize(
            self.repository, self.package_dir, os.path.basename(self.package_dir)
        )
        results = manage.list_system_reports(
            self.repository, os.path.basename(self.package_dir)
        )
        self.assertDictEqual(results, {})


if __name__ == "__main__":
    unittest.main()



"""
Command line interface for the Humbug infestor.
"""
import argparse
import os
import sys

from . import config, manage


def handle_config_init(args: argparse.Namespace) -> None:
    config.initialize(
        args.repository,
        args.python_root,
        args.name,
        args.relative_imports,
        args.reporter_token,
    )
    print("Infestor has been initialized for your repository.\n")

    print(
        "The Bugout team would like to collect crash reports as well as some basic, anonymous "
        "information about your system when you run infestor.\nThis will help us improve the "
        "infestor experience for everyone.\nWe request that you opt into reporting by setting the "
        "environment variable:\n"
        "\tINFESTOR_REPORTING_ENABLED=yes\n"
    )
    print("In bash or zsh:\n\t$ export INFESTOR_REPORTING_ENABLED=yes")
    print("In fish:\n\t$ set -x INFESTOR_REPORTING_ENABLED yes")
    print("On Windows (cmd or powershell):\n\t$ set INFESTOR_REPORTING_ENABLED=yes\n")

    if args.reporter_token is None:
        print(
            "It looks like you have not configured Infestor with a Bugout reporter token."
        )
        print(
            "Generate a reporter token by adding an integration to your team at https://bugout.dev/account/teams."
        )
        print("Once you have generated a token, run:")
        print("\t$ infestor token <token>\n")


def handle_config_validate(args: argparse.Namespace) -> None:
    config_file = config.default_config_file(args.repository)
    config.load_config(config_file, print_warnings=True)


def handle_config_token(args: argparse.Namespace) -> None:
    config_file = config.default_config_file(args.repository)
    config_object = config.set_reporter_token(
        config_file,
        config.python_root_relative_to_repository_root(
            args.repository, args.python_root
        ),
        args.token,
    )
    print(config_object)


def handle_reporter_add(args: argparse.Namespace) -> None:
    manage.add_reporter(args.repository, args.python_root, args.reporter_filepath)


def handle_system_report_add(args: argparse.Namespace) -> None:
    manage.add_system_report(args.repository, args.python_root, args.submodule)


def handle_system_report_list(args: argparse.Namespace) -> None:
    results = manage.list_system_reports(args.repository, args.python_root)
    for filepath, calls in results.items():
        print(f"Lines in {filepath}:")
        for report_call in calls:
            print(f"\t- {report_call.lineno}")


def handle_system_report_remove(args: argparse.Namespace) -> None:
    manage.remove_system_report(args.repository, args.python_root, args.submodule)


def generate_argument_parser() -> argparse.ArgumentParser:
    current_working_directory = os.getcwd()

    parser = argparse.ArgumentParser(
        description="Infestor: Manage Humbug instrumentation of your Python code base"
    )
    parser.set_defaults(func=lambda _: parser.print_help())
    subcommands = parser.add_subparsers()

    config_parser = subcommands.add_parser(
        "config", description="Manage infestor configuration"
    )
    config_parser.set_defaults(func=lambda _: config_parser.print_help())
    config_subcommands = config_parser.add_subparsers()

    config_init_parser = config_subcommands.add_parser(
        "init", description="Initialize an Infestor integration in a project"
    )
    config_init_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    config_init_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help=(
            "Root directory for Python code/module in the repository. If you are integrating with "
            "a module, this will be the highest-level directory with an __init__.py file in it."
        ),
    )
    config_init_parser.add_argument(
        "-n",
        "--name",
        required=True,
        help="Name of project (to identify integration)",
    )
    config_init_parser.add_argument(
        "--relative-imports",
        action="store_true",
        help="Set this flags if infestor should add relative imports.",
    )
    config_init_parser.add_argument(
        "-t",
        "--reporter-token",
        default=None,
        help="Bugout reporter token. Get one by setting up an integration at https://bugout.dev/account/teams",
    )
    config_init_parser.set_defaults(func=handle_config_init)

    config_validate_parser = config_subcommands.add_parser(
        "validate", description="Validate an Infestor configuration"
    )
    config_validate_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    config_validate_parser.set_defaults(func=handle_config_validate)

    config_token_parser = config_subcommands.add_parser(
        "token", description="Set a Humbug token for an Infestor integration"
    )
    config_token_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    config_token_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help="Root directory for Python code/module you want to register a token for (this is the relevant key in infestor.json)",
    )
    config_token_parser.add_argument(
        "token", help="Reporting token generated from https://bugout.dev/account/teams"
    )
    config_token_parser.set_defaults(func=handle_config_token)

    reporter_parser = subcommands.add_parser(
        "reporter", description="Manage Humbug reporters in a code base"
    )
    reporter_parser.set_defaults(func=lambda _: reporter_parser.print_help())
    reporter_subcommands = reporter_parser.add_subparsers()

    reporter_add_parser = reporter_subcommands.add_parser(
        "add", description="Adds a Humbug reporter to a Python package"
    )
    reporter_add_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    reporter_add_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help="Root directory for Python code/module you want to setup reporting for (this is the relevant key in infestor.json)",
    )
    reporter_add_parser.add_argument(
        "-o",
        "--reporter-filepath",
        required=False,
        default=manage.DEFAULT_REPORTER_FILENAME,
        help=f"Path (relative to Python root) at which we should set up the reporter integration (default: {manage.DEFAULT_REPORTER_FILENAME})",
    )
    reporter_add_parser.set_defaults(func=handle_reporter_add)

    system_report_parser = subcommands.add_parser(
        "system-report", description="Manage Humbug system reporting in a code base"
    )
    system_report_parser.set_defaults(func=lambda _: system_report_parser.print_help())
    system_report_subcommands = system_report_parser.add_subparsers()

    system_report_add_parser = system_report_subcommands.add_parser(
        "add",
        description="Adds reporting code to a given module",
    )
    system_report_add_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    system_report_add_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help="Root directory for Python code/module you want to setup reporting for (this is the relevant key in infestor.json)",
    )
    system_report_add_parser.add_argument(
        "-m",
        "--submodule",
        default=None,
        help="Path (relative to Python root) to submodule in which to fire off a system report",
    )
    system_report_add_parser.set_defaults(func=handle_system_report_add)

    system_report_list_parser = system_report_subcommands.add_parser(
        "list",
        description="Adds reporting code to a given module",
    )
    system_report_list_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    system_report_list_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help="Root directory for Python code/module you want to setup reporting for (this is the relevant key in infestor.json)",
    )
    system_report_list_parser.set_defaults(func=handle_system_report_list)

    system_report_remove_parser = system_report_subcommands.add_parser(
        "remove", description="Removes reporting code from a given module"
    )
    system_report_remove_parser.add_argument(
        "-r",
        "--repository",
        default=current_working_directory,
        help=f"Path to git repository containing your code base (default: {current_working_directory})",
    )
    system_report_remove_parser.add_argument(
        "-P",
        "--python-root",
        required=True,
        help="Root directory for Python code/module you want to setup reporting for (this is the relevant key in infestor.json)",
    )
    system_report_remove_parser.add_argument(
        "-m",
        "--submodule",
        default=None,
        help="Path (relative to Python root) to submodule in which to fire off a system report",
    )
    system_report_remove_parser.set_defaults(func=handle_system_report_remove)

    return parser


def main() -> None:
    parser = generate_argument_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""
These are the tools infestor uses to set up a code base for automatic Humbug instrumentation.
"""
from dataclasses import asdict, dataclass
import json
import os
from typing import Any, Dict, List, Optional, Tuple

from atomicwrites import atomic_write

from .version import INFESTOR_VERSION


CONFIG_FILENAME = "infestor.json"


@dataclass
class InfestorConfiguration:
    python_root: str
    project_name: str
    relative_imports: bool = False
    reporter_token: Optional[str] = None
    # reporter_filepath should be a path relative to python_root
    reporter_filepath: Optional[str] = None


REPORTER_TOKEN_KEY = "reporter_token"
PYTHON_ROOT_KEY = "python_root"
PROJECT_NAME_KEY = "project_name"
RELATIVE_IMPORTS_KEY = "relative_imports"
REPORTER_FILEPATH_KEY = "reporter_filepath"


class ConfigurationError(Exception):
    """
    Raised if there is an issue with an infestor configuration file.
    """

    pass


def parse_config(
    raw_config: Dict[str, Dict[str, Any]]
) -> Tuple[Dict[str, InfestorConfiguration], List[str], List[str]]:
    """
    Checks if the given configuration is valid. If it is valid, returns (True, []) else returns
    (False, [<warnings>, ...], [<error messages>, ...]).
    """
    warn_messages: List[str] = []
    error_messages: List[str] = []

    parsed_config: Dict[str, InfestorConfiguration] = {}

    for key_path, subconfiguration in raw_config.items():
        python_root: Optional[str] = subconfiguration.get(PYTHON_ROOT_KEY)
        if python_root is None:
            error_messages.append(f"{key_path}: No Python root directory specified")
            python_root = ""
        elif python_root != key_path:
            error_messages.append(
                f"{key_path}: Python root directory differs from the configuration key"
            )

        project_name: Optional[str] = subconfiguration.get(PROJECT_NAME_KEY)
        if project_name is None:
            error_messages.append(f"{key_path}: No project name specified")
            project_name = ""

        relative_imports: Optional[bool] = subconfiguration.get(RELATIVE_IMPORTS_KEY)
        if relative_imports is None:
            error_messages.append(
                f"{key_path}: Configuration does not specify whether or not to use relative imports"
            )
            relative_imports = False

        reporter_token: Optional[str] = subconfiguration.get(REPORTER_TOKEN_KEY)
        if reporter_token is None:
            warn_messages.append(f"{key_path}: No reporter token found")

        reporter_filepath: Optional[str] = subconfiguration.get(REPORTER_FILEPATH_KEY)
        if reporter_filepath is None:
            warn_messages.append(f"{key_path}: No reporter filepath found")

        if not error_messages:
            infestor_configuration = InfestorConfiguration(
                python_root=python_root,
                project_name=project_name,
                relative_imports=relative_imports,
                reporter_token=reporter_token,
                reporter_filepath=reporter_filepath,
            )
            parsed_config[infestor_configuration.python_root] = infestor_configuration

    return (parsed_config, warn_messages, error_messages)


def load_config(
    config_file: str, print_warnings: bool = False
) -> Dict[str, InfestorConfiguration]:
    """
    Loads an infestor configuration from file and validates it.
    """
    try:
        with open(config_file, "r") as ifp:
            raw_config = json.load(ifp)
    except:
        raise ConfigurationError(f"Could not read configuration: {config_file}")

    configuration, warnings, errors = parse_config(raw_config)

    if print_warnings:
        warning_items = "\n".join([f"- {warning}" for warning in warnings])
        if warnings:
            print(
                f"Warnings when loading configuration file ({config_file}):\n{warning_items}"
            )

    if errors:
        error_items = "\n".join([f"- {error}" for error in errors])
        error_message = (
            f"Errors loading configuration file ({config_file}):\n{error_items}"
        )
        raise ConfigurationError(error_message)

    return configuration


def save_config(config_file: str, configuration: InfestorConfiguration) -> None:
    result_configuration: Dict[str, Any] = {}
    if os.path.exists(config_file):
        existing_configuration = load_config(config_file, print_warnings=False)
        for python_root in existing_configuration:
            result_configuration[python_root] = asdict(
                existing_configuration[python_root]
            )

    result_configuration[configuration.python_root] = asdict(configuration)
    with atomic_write(config_file, overwrite=True) as ofp:
        json.dump(result_configuration, ofp)


def default_config_file(root_directory) -> str:
    config_file = os.path.join(root_directory, CONFIG_FILENAME)

    return config_file


def set_reporter_token(
    config_file: str, python_root: str, reporter_token: str
) -> Dict[str, InfestorConfiguration]:
    config = load_config(config_file)
    config[python_root].reporter_token = reporter_token
    save_config(config_file, config[python_root])
    return config


def initialize(
    repository: str,
    python_root: str,
    project_name: str,
    relative_imports: bool = False,
    reporter_token: Optional[str] = None,
) -> None:
    """
    Initialize infestor in a given project.
    """
    config_file = default_config_file(repository)

    configuration = InfestorConfiguration(
        python_root=python_root_relative_to_repository_root(repository, python_root),
        project_name=project_name,
        relative_imports=relative_imports,
        reporter_token=reporter_token,
    )
    save_config(config_file, configuration)


def python_root_relative_to_repository_root(repository: str, python_root: str) -> str:
    absolute_python_root = os.path.abspath(python_root)
    relative_python_root = os.path.relpath(absolute_python_root, repository)
    return relative_python_root



import os
from setuptools import find_packages, setup

from infestor.version import INFESTOR_VERSION

long_description = ""
README_PATH = os.path.join(os.path.dirname(__file__), "..", "README.md")
with open(README_PATH) as ifp:
    long_description = ifp.read()

setup(
    name="bugout-infestor",
    version=INFESTOR_VERSION,
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=["atomicwrites", "humbug", "pygit2"],
    extras_require={
        "dev": ["black", "mypy", "wheel"],
        "distribute": ["setuptools", "twine", "wheel"],
    },
    description="Humbug Infestor: Manage Humbug reporting over your code base",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Bugout.dev",
    author_email="engineering@bugout.dev",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries",
    ],
    url="https://github.com/bugout-dev/infestor",
    entry_points={"console_scripts": ["infestor=infestor.cli:main"]},
)



import ast
import importlib.util
import logging
import os
from typing import Any, cast, Dict, List, Optional, Tuple, Sequence

from .config import (
    default_config_file,
    load_config,
    save_config,
    python_root_relative_to_repository_root,
)

DEFAULT_REPORTER_FILENAME = "report.py"
REPORTER_FILE_TEMPLATE: Optional[str] = None
TEMPLATE_FILEPATH = os.path.join(os.path.dirname(__file__), "report.py.template")
try:
    with open(TEMPLATE_FILEPATH, "r") as ifp:
        REPORTER_FILE_TEMPLATE = ifp.read()
except Exception as e:
    logging.warn(f"WARNING: Could not load reporter template from {TEMPLATE_FILEPATH}:")
    logging.warn(e)


class GenerateReporterError(Exception):
    pass


class CallVisitor(ast.NodeVisitor):
    def __init__(self):
        self.calls: List[ast.Call] = []

    def visit_Call(self, node: ast.Call) -> Any:
        self.calls.append(node)


def python_files(repository: str, python_root: str) -> Sequence[str]:
    results: List[str] = []
    root = os.path.join(repository, python_root)
    if os.path.isfile(root):
        return [root]

    for dirpath, _, filenames in os.walk(root, topdown=True):
        results.extend(
            [
                os.path.join(dirpath, filename)
                for filename in filenames
                if filename.endswith(".py")
            ]
        )

    return results


def add_system_report(
    repository: str, python_root: str, submodule_path: Optional[str] = None
) -> None:
    """
    Args:
    1. repository - Path to repository in which Infestor has been set up
    2. python_root - Path (relative to repository) of Python package to work with (used to parse config)
    3. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    """
    config_file = default_config_file(repository)
    configuration = load_config(config_file).get(python_root)
    if configuration is None:
        raise GenerateReporterError(
            f"Could not find Python root ({python_root}) in configuration file ({config_file})"
        )

    if configuration.reporter_filepath is None:
        raise GenerateReporterError(
            f"No reporter defined for project. Try running:\n\t$ infestor -r {repository} generate setup -P {python_root} -o report.py"
        )
    reporter_filepath = os.path.join(
        repository, python_root, configuration.reporter_filepath
    )

    target_file = submodule_path
    if target_file is None:
        target_file = os.path.join(repository, python_root)
        if os.path.isdir(target_file):
            target_file = os.path.join(target_file, "__init__.py")

    if not os.path.exists(target_file):
        with open(target_file, "w") as ofp:
            ofp.write("")

    module: Optional[ast.Module] = None
    with open(target_file, "r") as ifp:
        module = ast.parse(ifp.read())

    last_import_line_number = 0
    for statement in module.body:
        if isinstance(statement, ast.Import) or isinstance(statement, ast.ImportFrom):
            last_import_line_number = statement.lineno

    # TODO(zomglings): Create an AST node which imports the reporter and runs reporter.system_report()
    path_to_reporter_file = os.path.relpath(
        os.path.join(repository, python_root, reporter_filepath),
        os.path.dirname(target_file),
    )
    path_components: List[str] = []
    current_path = path_to_reporter_file
    while current_path:
        current_path, base = os.path.split(current_path)
        if base == os.path.basename(reporter_filepath):
            base, _ = os.path.splitext(base)
        path_components = [base] + path_components

    source_lines: List[str] = []
    with open(target_file, "r") as ifp:
        for line in ifp:
            source_lines.append(line)

    new_code = ""
    if not configuration.relative_imports:
        path_components = [os.path.basename(python_root)] + path_components
        name = ".".join(path_components)
        new_code = f"import {name}\n{name}.reporter.system_report()\n"
    else:
        name = ".".join(path_components)
        new_code = f"from .{name} import reporter\nreporter.system_report()\n"

    source_lines = (
        source_lines[:last_import_line_number]
        + [new_code]
        + source_lines[last_import_line_number:]
    )

    with open(target_file, "w") as ofp:
        for line in source_lines:
            ofp.write(line)


def list_reporter_imports(
    repository: str, python_root: str, candidate_files: Optional[Sequence[str]] = None
) -> Dict[str, ast.Module]:
    """
    Args:
    1. repository - Path to repository in which Infestor has been set up
    2. python_root - Path (relative to repository) of Python package to work with (used to parse config)
    3. candidate_files - Optional list of files to restrict analysis to
    """
    results: Dict[str, ast.Module] = {}

    config_file = default_config_file(repository)
    configuration = load_config(config_file).get(python_root)
    if configuration is None:
        raise GenerateReporterError(
            f"Could not find Python root ({python_root}) in configuration file ({config_file})"
        )

    if configuration.reporter_filepath is None:
        # No infestor-managed reporter file, so just return quietly
        return results

    # Until the end of the loop, this is reversed
    reporter_filepath = os.path.join(python_root, configuration.reporter_filepath)
    dirname, basename = os.path.split(reporter_filepath)
    base_module, _ = os.path.splitext(basename)
    reporter_module_components: List[str] = [base_module]
    while True:
        dirname, basename = os.path.split(dirname)
        if basename == "":
            break
        reporter_module_components.append(basename)
        if dirname == "":
            break
    reporter_module_components.reverse()
    reporter_module = ".".join(reporter_module_components)

    if candidate_files is None:
        candidate_files = python_files(repository, python_root)

    for candidate_file in candidate_files:
        module: Optional[ast.Module] = None
        with open(candidate_file, "r") as ifp:
            module = ast.parse(ifp.read())

        for statement in module.body:
            if isinstance(statement, ast.Import):
                for name in statement.names:
                    if name.name == reporter_module:
                        results[candidate_file] = module
            elif isinstance(statement, ast.ImportFrom):
                module_name = f"{'.'*statement.level}{statement.module}"
                qualified_module_name = importlib.util.resolve_name(
                    module_name, python_root
                )
                if qualified_module_name == reporter_module:
                    results[candidate_file] = module

    return results


def list_system_reports(
    repository: str, python_root: str, candidate_files: Optional[Sequence[str]] = None
) -> Dict[str, List[ast.Call]]:
    """
    Args:
    1. repository - Path to repository in which Infestor has been set up
    2. python_root - Path (relative to repository) of Python package to work with (used to parse config)
    3. candidate_files - Optional list of files to restrict analysis to
    """
    results: Dict[str, List[ast.Call]] = {}
    files_with_reporter = list_reporter_imports(
        repository, python_root, candidate_files
    )

    call_logger = CallVisitor()
    for filepath, file_ast in files_with_reporter.items():
        system_reports: List[ast.Call] = []
        call_logger.calls = []
        call_logger.visit(file_ast)

        for call_object in call_logger.calls:
            if (
                isinstance(call_object.func, ast.Name)
                and call_object.func.id == "system_report"
            ):
                system_reports.append(call_object)
            elif (
                isinstance(call_object.func, ast.Attribute)
                and call_object.func.attr == "system_report"
            ):
                system_reports.append(call_object)

        if system_reports:
            results[filepath] = system_reports

    return results


def remove_system_report(
    repository: str, python_root: str, submodule_path: Optional[str] = None
) -> None:
    """
    Args:
    1. repository - Path to repository in which Infestor has been set up
    2. python_root - Path (relative to repository) of Python package to work with (used to parse config)
    3. submodule_path: Path (relative to python_root) of file in which we want to add a sytem_report
    """
    candidate_files: Sequence[Optional[str]] = [submodule_path]
    if submodule_path is None:
        candidate_files = python_files(repository, python_root)
    candidate_files = cast(Sequence[str], candidate_files)

    system_report_calls = list_system_reports(repository, python_root, candidate_files)

    for filepath, calls in system_report_calls.items():
        deletions: List[Tuple[int, Optional[int]]] = [
            (report_call.lineno, report_call.end_lineno) for report_call in calls
        ]

        if deletions:
            current_deletion = 0
            new_lines = []
            with open(filepath, "r") as ifp:
                for j, line in enumerate(ifp):
                    i = j + 1
                    if i >= deletions[current_deletion][0] and (
                        deletions[current_deletion][1] is None
                        or i <= cast(int, deletions[current_deletion][1])
                    ):
                        continue
                    else:
                        new_lines.append(line)
                    if i == deletions[current_deletion][
                        1
                    ] and current_deletion + 1 < len(deletions):
                        current_deletion += 1

            with open(filepath, "w") as ofp:
                ofp.write("".join(new_lines))


def add_reporter(
    repository: str,
    python_root: str,
    reporter_filepath: Optional[str] = None,
) -> None:
    if REPORTER_FILE_TEMPLATE is None:
        raise GenerateReporterError("Could not load reporter template file")

    config_file = default_config_file(repository)
    configurations_by_python_root = load_config(config_file)

    normalized_python_root = python_root_relative_to_repository_root(
        repository, python_root
    )

    configuration = configurations_by_python_root.get(normalized_python_root)
    if configuration is None:
        raise GenerateReporterError(
            f"Could not find configuration for python root ({python_root}) in config file ({config_file})"
        )

    if reporter_filepath is None:
        if configuration.reporter_filepath is not None:
            reporter_filepath = configuration.reporter_filepath
        else:
            reporter_filepath = DEFAULT_REPORTER_FILENAME
    else:
        if (
            configuration.reporter_filepath is not None
            and configuration.reporter_filepath != reporter_filepath
        ):
            raise GenerateReporterError(
                f"Configuration expects reporter to be set up at a different file than the one specified; specified={reporter_filepath}, expected={configuration.reporter_filepath}"
            )

    reporter_filepath_full = os.path.join(
        repository, configuration.python_root, reporter_filepath
    )
    if os.path.exists(reporter_filepath_full):
        raise GenerateReporterError(
            f"Object already exists at desired reporter filepath: {reporter_filepath_full}"
        )

    if configuration.reporter_token is None:
        raise GenerateReporterError("No reporter token was specified in configuration")

    contents = REPORTER_FILE_TEMPLATE.format(
        project_name=configuration.project_name,
        reporter_token=configuration.reporter_token,
    )
    with open(reporter_filepath_full, "w") as ofp:
        ofp.write(contents)

    configuration.reporter_filepath = reporter_filepath
    save_config(config_file, configuration)



# Try to set up reporting. If it doesn't work because imported dependencies are not present in
# the user's environment, fail quietly.
try:
    from .report import infestor_reporter, infestor_tags

    infestor_reporter.system_report(tags=infestor_tags)
    infestor_reporter.setup_excepthook(tags=infestor_tags)
except Exception:
    pass



import json
import os
import shutil
import tempfile
import unittest

from . import config


class TestParseConfig(unittest.TestCase):
    maxDiff = None

    # TODO(zomglings): Test reporter_filepath invalidities.

    def test_empty_config(self):
        raw_config = {}
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 0)
        self.assertDictEqual(configuration, {})

    def test_config_with_key_different_from_python_root(self):
        raw_config = {
            "./lol": {
                config.PYTHON_ROOT_KEY: "./rofl",
                config.PROJECT_NAME_KEY: "lol",
                config.RELATIVE_IMPORTS_KEY: False,
                config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
                config.REPORTER_FILEPATH_KEY: "report.py",
            }
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 1)
        self.assertDictEqual(configuration, {})

    def test_config_with_no_project_name(self):
        raw_config = {
            "./rofl": {
                config.PYTHON_ROOT_KEY: "./rofl",
                config.RELATIVE_IMPORTS_KEY: False,
                config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
                config.REPORTER_FILEPATH_KEY: "report.py",
            }
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 1)
        self.assertDictEqual(configuration, {})

    def test_config_with_no_relative_imports(self):
        raw_config = {
            "./rofl": {
                config.PYTHON_ROOT_KEY: "./rofl",
                config.PROJECT_NAME_KEY: "rofl",
                config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
                config.REPORTER_FILEPATH_KEY: "report.py",
            }
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 1)
        self.assertDictEqual(configuration, {})

    def test_config_with_no_python_root_or_project_name(self):
        raw_config = {
            "./rofl": {
                config.RELATIVE_IMPORTS_KEY: False,
                config.REPORTER_TOKEN_KEY: "335da960-2dc6-48b3-97a5-c23ac1495e7d",
                config.REPORTER_FILEPATH_KEY: "report.py",
            }
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 2)
        self.assertDictEqual(configuration, {})

    def test_single_valid_config(self):
        infestor_configuration = config.InfestorConfiguration(
            python_root="./lol",
            project_name="lol",
            relative_imports=False,
            reporter_token="335da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )
        raw_config = {
            infestor_configuration.python_root: {
                config.PYTHON_ROOT_KEY: infestor_configuration.python_root,
                config.PROJECT_NAME_KEY: infestor_configuration.project_name,
                config.RELATIVE_IMPORTS_KEY: infestor_configuration.relative_imports,
                config.REPORTER_TOKEN_KEY: infestor_configuration.reporter_token,
                config.REPORTER_FILEPATH_KEY: infestor_configuration.reporter_filepath,
            }
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 0)
        self.assertDictEqual(
            configuration, {infestor_configuration.python_root: infestor_configuration}
        )

    def test_two_valid_configs(self):
        infestor_configuration_0 = config.InfestorConfiguration(
            python_root="./lol",
            project_name="lol",
            relative_imports=False,
            reporter_token="335da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )
        infestor_configuration_1 = config.InfestorConfiguration(
            python_root="./rofl",
            project_name="rofl",
            relative_imports=False,
            reporter_token="435da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )

        raw_config = {
            infestor_configuration_0.python_root: {
                config.PYTHON_ROOT_KEY: infestor_configuration_0.python_root,
                config.PROJECT_NAME_KEY: infestor_configuration_0.project_name,
                config.RELATIVE_IMPORTS_KEY: infestor_configuration_0.relative_imports,
                config.REPORTER_TOKEN_KEY: infestor_configuration_0.reporter_token,
                config.REPORTER_FILEPATH_KEY: infestor_configuration_0.reporter_filepath,
            },
            infestor_configuration_1.python_root: {
                config.PYTHON_ROOT_KEY: infestor_configuration_1.python_root,
                config.PROJECT_NAME_KEY: infestor_configuration_1.project_name,
                config.RELATIVE_IMPORTS_KEY: infestor_configuration_1.relative_imports,
                config.REPORTER_TOKEN_KEY: infestor_configuration_1.reporter_token,
                config.REPORTER_FILEPATH_KEY: infestor_configuration_1.reporter_filepath,
            },
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 0)
        self.assertDictEqual(
            configuration,
            {
                infestor_configuration_0.python_root: infestor_configuration_0,
                infestor_configuration_1.python_root: infestor_configuration_1,
            },
        )

    def test_valid_and_invalid_configs(self):
        infestor_configuration_0 = config.InfestorConfiguration(
            python_root="./lol",
            project_name="lol",
            relative_imports=False,
            reporter_token="335da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )
        infestor_configuration_1 = config.InfestorConfiguration(
            python_root="./rofl",
            project_name="rofl",
            relative_imports=False,
            reporter_token="435da960-2dc6-48b3-97a5-c23ac1495e7d",
            reporter_filepath="report.py",
        )

        raw_config = {
            infestor_configuration_0.python_root: {
                config.PYTHON_ROOT_KEY: infestor_configuration_0.python_root,
                config.PROJECT_NAME_KEY: infestor_configuration_0.project_name,
                config.RELATIVE_IMPORTS_KEY: infestor_configuration_0.relative_imports,
                config.REPORTER_TOKEN_KEY: infestor_configuration_0.reporter_token,
                config.REPORTER_FILEPATH_KEY: infestor_configuration_0.reporter_filepath,
            },
            f"{infestor_configuration_1.python_root}/": {
                config.PYTHON_ROOT_KEY: infestor_configuration_1.python_root,
                config.PROJECT_NAME_KEY: infestor_configuration_1.project_name,
                config.REPORTER_TOKEN_KEY: infestor_configuration_1.reporter_token,
                config.REPORTER_FILEPATH_KEY: infestor_configuration_1.reporter_filepath,
            },
        }
        configuration, warnings, errors = config.parse_config(raw_config)
        self.assertEqual(len(warnings), 0)
        self.assertEqual(len(errors), 2)
        self.assertDictEqual(
            configuration,
            {
                infestor_configuration_0.python_root: infestor_configuration_0,
            },
        )


class TestInit(unittest.TestCase):
    def setUp(self):
        self.repository = tempfile.mkdtemp()
        self.project_name = "my-awesome-python-project"

    def tearDown(self):
        shutil.rmtree(self.repository)

    def test_initialize_infestor_once(self):
        config.initialize(self.repository, self.repository, self.project_name)
        config_file = config.default_config_file(self.repository)
        self.assertTrue(os.path.isfile(config_file))
        with open(config_file, "r") as ifp:
            configuration = json.load(ifp)

        configuration_key = config.python_root_relative_to_repository_root(
            self.repository,
            self.repository,
        )
        self.assertIsNotNone(configuration.get(configuration_key))

        self.assertEqual(len(configuration), 1)

    def test_initialize_infestor_twice(self):
        package_0 = os.path.join(self.repository, "lol")
        os.mkdir(package_0)
        package_1 = os.path.join(self.repository, "rofl")
        os.mkdir(package_1)

        config.initialize(self.repository, package_0, "lol")
        config.initialize(self.repository, package_1, "rofl")

        config_file = config.default_config_file(self.repository)
        self.assertTrue(os.path.isfile(config_file))
        with open(config_file, "r") as ifp:
            configuration = json.load(ifp)

        self.assertIsNotNone(configuration.get("lol"))
        self.assertIsNotNone(configuration.get("rofl"))

        self.assertEqual(len(configuration), 2)

    def test_initialize_infestor_twice_with_same_python_root(self):
        package_0 = os.path.join(self.repository, "lol")
        os.mkdir(package_0)
        config.initialize(self.repository, package_0, "lol", relative_imports=False)
        config.initialize(self.repository, package_0, "lol", relative_imports=True)

        config_file = config.default_config_file(self.repository)
        self.assertTrue(os.path.isfile(config_file))
        with open(config_file, "r") as ifp:
            configuration = json.load(ifp)

        self.assertIsNotNone(configuration.get("lol"))
        self.assertEqual(len(configuration), 1)
        self.assertTrue(configuration["lol"][config.RELATIVE_IMPORTS_KEY])


if __name__ == "__main__":
    unittest.main()
